import socket
import types
import string
import BufferedSocket

BLOCKSIZE = 1024			# size of recv blocks
DEBUG = 0

class Client(BufferedSocket.BufferedSocket):
    """An IRC client connection to a server

    This isn't going to give you any special ability to speak IRC, you
    still have to know how to speak it.  This just makes it easier.  For
    making things really nice, check out irc_interface.py.

    """

    def __init__(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	self.__name = None
	BufferedSocket.BufferedSocket.__init__(self, sock)


    def _set_name(self, name):
        """Set the nickname of this connection"""
        
	self.__name = name


    def name(self):
        """Return our name"""
        
	return self.__name

    
    def fileno(self):
	"""Return a file number corresponding to this object's socket.

	This is primarily useful for passing the object to
	select.select(), although you could use it for linking in with
	other programs or anything else you need a file descriptor for.

	"""
	return self.sock.fileno()


    def connect(self, server, port, nick, gecos):
	"""Establish an IRC connection

	Connect to server on port, using nick for the initial nickname,
	and user for the RL-username.  This is merely a convenience
	function--if you're writing an advanced client you will probably
	want to do these things yourself, or overload this.
	
	"""

	self._set_name("%s:%d" % (server, port))
       
	self.sock.connect((server, port))
	self.write(['NICK', nick])
	self.write(['USER', nick, '+iw', nick], gecos)
	

    def write(self, args, *text):
	"""Send out an IRC command

	This function helps to prevent you from shooting yourself in the
	foot, by forcing you to send commands that are in a valid format
	(although it doesn't check the validity of the actual commands).
	
	As we all know, IRC commands take the form

	  :COMMAND ARG1 ARG2 ARG3 ... :text string
    
	where "text string" is optional.  Well, that's exactly how this
	function works.  Args is a list of length at least one, and text
	string can be any number of strings.  You can call the function
	like this:

	  write(["PRIVMSG", nick], "Hello 12")

	or like this:

	  write(["PRIVMSG", nick], "Hello", "12")

        or even like this:

	  write(["PRIVMSG", nick], "Hello", 12)

	And you'll get the same result with all three.

	"""
       
	if (type(args) == types.StringType):
	    cmd = args
	else:
	    cmd = string.join(args)

	cmdstr = cmd
	
	if (text):
	    txt = ''
	    for t in (text):
		if (type(t) == types.StringType):
		    txt = txt + ' ' + t
		elif (type(t) in (types.ListType, types.TupleType)):
		    for i in (t):
			try:
			    txt = string.join([txt, i])
			except TypeError:
			    txt = string.join([txt, repr(i)])
		else:
		    txt = string.join([txt, repr(t)])

	    txt = txt[1:]

	    cmdstr = "%s :%s" % (cmdstr, txt)

	if DEBUG:
	    print "-> " + cmdstr
	 
	self.sock.send(cmdstr + '\n')


    def parse_line(self, line):
	"""Parse a server-provided line

	This does all the magic of parsing those ill-formatted IRC
	messages.  It will also decide if a PRIVMSG or NOTICE is using
	CTCP (the client-to-client protocol, which by convention is any
	of the above messages with ^A on both ends of the text.

	This function goes on to invoke self.eval_triggers on the parsed
	data like this:

	  self.eval_triggers(operation, arguments, text)

	where operation and text are strings, and arguments is a list.

	It returns the same tuple (op, args, text).

	"""
	if (line[0] == ':'):
	    with_uname = 1
	    line = line [1:]
	else:
	    with_uname = 0

	try:
	    [args, text] = string.split(line, ' :', 1)
	    args = string.split(args)
	except ValueError:
	    args = string.split(line)
	    text = ''

	if (with_uname != 1):
	    op = args[0]
	elif ((args[1] in ["PRIVMSG", "NOTICE"]) and
	      ((text[0] == '\001') and (text[-1] == '\001'))):
	    op = "C" + args[1]
	    text = text[1:-1]
	else:
	    op = args[1]

	if DEBUG:
	    print "<-", op, args, text

	self.eval_triggers(op, args, text)

	return (op, args, text)


    def unhandled(self, op, args, text):
	"""Default for dealing with unhandled server messages

	This is something you may or may not want to override.

	"""

	pass

    def eval_triggers(self, op, args, text):
	"""Take action on a server message

	Right now, just invokes

	  self.do_[op](args, text)

	where [op] is the operation passed in.

	This is a good method to overload if you want a really advanced
	client supporting bindings.

	"""
	try:
	    method = getattr(self, "do_" + lower(op))
	except AttributeError:
	    return self.unhandled(op, args, text)
	method(args, text)
      

    def do_ping(self, args, text):
	"""Reply to a server PING

	IRC servers send out perioding PINGs to make sure the client
	isn't comatose.  This method is provided as a convenient default
	method to reply to these PINGs so your client isn't suddenly
	disconnected with a "Ping Timeout" error.

	Overload away.

	"""

	self.write('PONG', text)
	    


##
## Miscellaneous IRC functions
##

ucletters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ[]\\"
lcletters = "abcdefghijklmnopqrstuvwxyz{}|"

ultrans = string.maketrans(ucletters, lcletters)
lutrans = string.maketrans(lcletters, ucletters)
casetrans = string.maketrans(string.join([ucletters, lcletters], ''),
			     string.join([lcletters, ucletters], ''))

def upper(s):
    """Convert a string to upper case

    Because IRC was developed in a nordic country, there are three extra
    letters.  In order to do case conversions properly, you need to use
    the IRC-specific functions.

    """

    return string.translate(s, lutrans)

def lower(s):
    """Convert a string to lower case

    Because IRC was developed in a nordic country, there are three extra
    letters.  In order to do case conversions properly, you need to use
    the IRC-specific functions.

    """

    return string.translate(s, ultrans)

def swapcase(s):
    """Invert a string's case

    Because IRC was developed in a nordic country, there are three extra
    letters.  In order to do case conversions properly, you need to use
    the IRC-specific functions.

    """

    return string.translate(s, casetrans)


def strcmp(s1, s2):
    """Case-insensitively compare two strings

    Because IRC was developed in a nordic country, there are three extra
    letters.  In order to do case-insensitive comparisons properly, you
    need to use this IRC-specific function.

    """

    return (lower(s1) == lower(s2))


def unpack_nuhost(nuhost):
    """Unpack nick!user@host

    Frequently, the first argument in a server message is in
    nick!user@host format.  You can just pass args[0] to this function
    and get back a tuple containing:

      (nick, user, host)

    """

    try:
	[nick, uhost] = string.split(nuhost, '!', 1)
	[user, host] = string.split(uhost, '@', 1)
    except ValueError:
	raise ValueError, "not in nick!user@host format"
    return (nick, user, host)
